@@ -1,7 +1,7 @@
 #define HOT_LRU 0
 #define WARM_LRU 64
 #define COLD_LRU 128
-#define NOEXP_LRU 192
+#define TEMP_LRU 192
 
 #define CLEAR_LRU(id) (id & ~(3<<6))
 
@@ -27,6 +27,8 @@ void do_item_linktail_q(item *it);
 void do_item_unlinktail_q(item *it);
 item *do_item_crawl_q(item *it);
 
+void *item_lru_bump_buf_create(void);
+
 /*@null@*/
 char *item_cachedump(const unsigned int slabs_clsid, const unsigned int limit, unsigned int *bytes);
 void item_stats(ADD_STAT add_stats, void *c);
@@ -42,7 +44,7 @@ void item_stats_sizes_add(item *it);
 void item_stats_sizes_remove(item *it);
 bool item_stats_sizes_status(void);
 
-item *do_item_get(const char *key, const size_t nkey, const uint32_t hv, conn *c);
+item *do_item_get(const char *key, const size_t nkey, const uint32_t hv, conn *c, const bool do_update);
 item *do_item_touch(const char *key, const size_t nkey, uint32_t exptime, const uint32_t hv, conn *c);
 void item_stats_reset(void);
 extern pthread_mutex_t lru_locks[POWER_LARGEST];
@@ -52,3 +54,5 @@ int stop_lru_maintainer_thread(void);
 int init_lru_maintainer(void);
 void lru_maintainer_pause(void);
 void lru_maintainer_resume(void);
+
+void *lru_bump_buf_create(void);
