@@ -7,6 +7,10 @@
 #include <poll.h>
 #include <ctype.h>
 
+#if defined(__sun)
+#include <atomic.h>
+#endif
+
 #include "memcached.h"
 #include "bipbuffer.h"
 
@@ -44,7 +48,10 @@ static const entry_details default_entries[] = {
     [LOGGER_ASCII_CMD] = {LOGGER_TEXT_ENTRY, 512, LOG_RAWCMDS, "<%d %s"},
     [LOGGER_EVICTION] = {LOGGER_EVICTION_ENTRY, 512, LOG_EVICTIONS, NULL},
     [LOGGER_ITEM_GET] = {LOGGER_ITEM_GET_ENTRY, 512, LOG_FETCHERS, NULL},
-    [LOGGER_ITEM_STORE] = {LOGGER_ITEM_STORE_ENTRY, 512, LOG_MUTATIONS, NULL}
+    [LOGGER_ITEM_STORE] = {LOGGER_ITEM_STORE_ENTRY, 512, LOG_MUTATIONS, NULL},
+    [LOGGER_CRAWLER_STATUS] = {LOGGER_TEXT_ENTRY, 512, LOG_SYSEVENTS,
+        "type=lru_crawler crawler=%d low_mark=%llu next_reclaims=%llu since_run=%u next_run=%d elapsed=%u examined=%llu reclaimed=%llu"
+    }
 };
 
 #define WATCHER_ALL -1
@@ -145,18 +152,18 @@ static int _logger_thread_parse_ise(logentry *e, char *scratch) {
     char keybuf[KEY_MAX_LENGTH * 3 + 1];
     struct logentry_item_store *le = (struct logentry_item_store *) e->data;
     const char * const status_map[] = {
-        "stored", "exists", "not_found", "too_large", "no_memory" };
+        "not_stored", "stored", "exists", "not_found", "too_large", "no_memory" };
     const char * const cmd_map[] = {
-        "add", "set", "replace", "append", "prepend", "cas" };
+        "null", "add", "set", "replace", "append", "prepend", "cas" };
 
     if (le->cmd <= 5)
         cmd = cmd_map[le->cmd];
 
     uriencode(le->key, keybuf, le->nkey, LOGGER_PARSE_SCRATCH);
     total = snprintf(scratch, LOGGER_PARSE_SCRATCH,
-            "ts=%d.%d gid=%llu type=item_store key=%s status=%s cmd=%s\n",
+            "ts=%d.%d gid=%llu type=item_store key=%s status=%s cmd=%s ttl=%u clsid=%u\n",
             (int)e->tv.tv_sec, (int)e->tv.tv_usec, (unsigned long long) e->gid,
-            keybuf, status_map[le->status], cmd);
+            keybuf, status_map[le->status], cmd, le->ttl, le->clsid);
     return total;
 }
 
@@ -169,9 +176,9 @@ static int _logger_thread_parse_ige(logentry *e, char *scratch) {
 
     uriencode(le->key, keybuf, le->nkey, LOGGER_PARSE_SCRATCH);
     total = snprintf(scratch, LOGGER_PARSE_SCRATCH,
-            "ts=%d.%d gid=%llu type=item_get key=%s status=%s\n",
+            "ts=%d.%d gid=%llu type=item_get key=%s status=%s clsid=%u\n",
             (int)e->tv.tv_sec, (int)e->tv.tv_usec, (unsigned long long) e->gid,
-            keybuf, was_found_map[le->was_found]);
+            keybuf, was_found_map[le->was_found], le->clsid);
     return total;
 }
 
@@ -181,10 +188,10 @@ static int _logger_thread_parse_ee(logentry *e, char *scratch) {
     struct logentry_eviction *le = (struct logentry_eviction *) e->data;
     uriencode(le->key, keybuf, le->nkey, LOGGER_PARSE_SCRATCH);
     total = snprintf(scratch, LOGGER_PARSE_SCRATCH,
-            "ts=%d.%d gid=%llu type=eviction key=%s fetch=%s ttl=%lld la=%d\n",
+            "ts=%d.%d gid=%llu type=eviction key=%s fetch=%s ttl=%lld la=%d clsid=%u\n",
             (int)e->tv.tv_sec, (int)e->tv.tv_usec, (unsigned long long) e->gid,
             keybuf, (le->it_flags & ITEM_FETCHED) ? "yes" : "no",
-            (long long int)le->exptime, le->latime);
+            (long long int)le->exptime, le->latime, le->clsid);
 
     return total;
 }
@@ -216,7 +223,7 @@ static enum logger_parse_entry_ret logger_thread_parse_entry(logentry *e, struct
         L_DEBUG("LOGGER: Failed to flatten log entry!\n");
         return LOGGER_PARSE_ENTRY_FAILED;
     } else {
-        *scratch_len = total + 1;
+        *scratch_len = total;
     }
 
     return LOGGER_PARSE_ENTRY_OK;
@@ -239,7 +246,7 @@ static void logger_thread_write_entry(logentry *e, struct logger_stats *ls,
         while (!w->failed_flush &&
                 (skip_scr = (char *) bipbuf_request(w->buf, scratch_len + 128)) == NULL) {
             if (logger_thread_poll_watchers(0, x) <= 0) {
-                L_DEBUG("LOGGER: Watcher had no free space for line of size (%d)\n", line_size);
+                L_DEBUG("LOGGER: Watcher had no free space for line of size (%d)\n", scratch_len + 128);
                 w->failed_flush = true;
             }
         }
@@ -259,7 +266,7 @@ static void logger_thread_write_entry(logentry *e, struct logger_stats *ls,
                 ls->watcher_skipped++;
                 continue;
             }
-            bipbuf_push(w->buf, total + 1);
+            bipbuf_push(w->buf, total);
             w->skipped = 0;
         }
         /* Can't fail because bipbuf_request succeeded. */
@@ -395,6 +402,7 @@ static int logger_thread_poll_watchers(int force_poll, int watcher) {
             char buf[1];
             int res = read(w->sfd, buf, 1);
             if (res == 0 || (res == -1 && (errno != EAGAIN && errno != EWOULDBLOCK))) {
+                L_DEBUG("LOGGER: watcher closed remotely\n");
                 logger_thread_close_watcher(w);
                 nfd++;
                 continue;
@@ -569,6 +577,7 @@ static void _logger_log_evictions(logentry *e, item *it) {
     le->latime = current_time - it->time;
     le->it_flags = it->it_flags;
     le->nkey = it->nkey;
+    le->clsid = ITEM_clsid(it);
     memcpy(le->key, ITEM_key(it), it->nkey);
     e->size = sizeof(struct logentry_eviction) + le->nkey;
 }
@@ -579,20 +588,28 @@ static void _logger_log_evictions(logentry *e, item *it) {
  * for more endpoints to be written before making it generic, though.
  * TODO: This and below should track and reprint the client fd.
  */
-static void _logger_log_item_get(logentry *e, const int was_found, const char *key, const int nkey) {
+static void _logger_log_item_get(logentry *e, const int was_found, const char *key,
+        const int nkey, const uint8_t clsid) {
     struct logentry_item_get *le = (struct logentry_item_get *) e->data;
     le->was_found = was_found;
     le->nkey = nkey;
+    le->clsid = clsid;
     memcpy(le->key, key, nkey);
     e->size = sizeof(struct logentry_item_get) + nkey;
 }
 
 static void _logger_log_item_store(logentry *e, const enum store_item_type status,
-        const int comm, char *key, const int nkey) {
+        const int comm, char *key, const int nkey, rel_time_t ttl, const uint8_t clsid) {
     struct logentry_item_store *le = (struct logentry_item_store *) e->data;
     le->status = status;
     le->cmd = comm;
     le->nkey = nkey;
+    le->clsid = clsid;
+    if (ttl != 0) {
+        le->ttl = ttl - current_time;
+    } else {
+        le->ttl = 0;
+    }
     memcpy(le->key, key, nkey);
     e->size = sizeof(struct logentry_item_store) + nkey;
 }
@@ -650,7 +667,8 @@ enum logger_ret_type logger_log(logger *l, const enum log_entry_type event, cons
             int was_found = va_arg(ap, int);
             char *key = va_arg(ap, char *);
             size_t nkey = va_arg(ap, size_t);
-            _logger_log_item_get(e, was_found, key, nkey);
+            uint8_t gclsid = va_arg(ap, int);
+            _logger_log_item_get(e, was_found, key, nkey, gclsid);
             va_end(ap);
             break;
         case LOGGER_ITEM_STORE_ENTRY:
@@ -659,7 +677,9 @@ enum logger_ret_type logger_log(logger *l, const enum log_entry_type event, cons
             int comm = va_arg(ap, int);
             char *skey = va_arg(ap, char *);
             size_t snkey = va_arg(ap, size_t);
-            _logger_log_item_store(e, status, comm, skey, snkey);
+            rel_time_t sttl = va_arg(ap, rel_time_t);
+            uint8_t sclsid = va_arg(ap, int);
+            _logger_log_item_store(e, status, comm, skey, snkey, sttl, sclsid);
             break;
     }
 
@@ -670,7 +690,8 @@ enum logger_ret_type logger_log(logger *l, const enum log_entry_type event, cons
         return LOGGER_RET_ERR;
     }
     l->written++;
-    L_DEBUG("LOGGER: Requested %d bytes, wrote %d bytes\n", reqlen, total + 1);
+    L_DEBUG("LOGGER: Requested %d bytes, wrote %lu bytes\n", reqlen,
+            (sizeof(logentry) + e->size));
 
     pthread_mutex_unlock(&l->mutex);
 
